==> v8.12.zip: prog8/codeGenIntermediate/src/prog8/codegen/intermediate/IRUnusedCodeRemover.kt <==
package prog8.codegen.intermediate

import prog8.code.core.IErrorReporter
import prog8.code.core.SourceCode.Companion.libraryFilePrefix
import prog8.intermediate.*


internal class IRUnusedCodeRemover(
    private val irprog: IRProgram,
    private val symbolTable: IRSymbolTable,
    private val errors: IErrorReporter
) {
    fun optimize(): Int {
        val allLabeledChunks = mutableMapOf<String, IRCodeChunkBase>()

        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.forEach { chunk ->
                chunk.label?.let { allLabeledChunks[it] = chunk }
            }
        }

        var numRemoved = removeSimpleUnlinked(allLabeledChunks) + removeUnreachable(allLabeledChunks)

        // remove empty subs
        irprog.blocks.forEach { block ->
            block.children.filterIsInstance<IRSubroutine>().reversed().forEach { sub ->
                if(sub.isEmpty()) {
                    if(!sub.position.file.startsWith(libraryFilePrefix)) {
                        errors.warn("unused subroutine ${sub.label}", sub.position)
                    }
                    block.children.remove(sub)
                    symbolTable.removeTree(sub.label)
                    numRemoved++
                }
            }
        }

        // remove empty blocks
        irprog.blocks.reversed().forEach { block ->
            if(block.isEmpty()) {
                irprog.blocks.remove(block)
                symbolTable.removeTree(block.label)
                numRemoved++
            }
        }

        return numRemoved
    }

    private fun removeUnreachable(allLabeledChunks: MutableMap<String, IRCodeChunkBase>): Int {
        val entrypointSub = irprog.blocks.single { it.label=="main" }.children.single { it is IRSubroutine && it.label=="main.start" }
        val reachable = mutableSetOf((entrypointSub as IRSubroutine).chunks.first())

        fun grow() {
            val new = mutableSetOf<IRCodeChunkBase>()
            reachable.forEach {
                it.next?.let { next -> new += next }
                it.instructions.forEach { instr ->
                    if (instr.branchTarget == null)
                        instr.labelSymbol?.let { label -> allLabeledChunks[label]?.let { chunk -> new += chunk } }
                    else
                        new += instr.branchTarget!!
                }
            }
            reachable += new
        }

        var previousCount = reachable.size
        while(true) {
            grow()
            if(reachable.size<=previousCount)
                break
            previousCount = reachable.size
        }

        return removeUnlinkedChunks(reachable)
    }

    private fun removeSimpleUnlinked(allLabeledChunks: Map<String, IRCodeChunkBase>): Int {
        val linkedChunks = mutableSetOf<IRCodeChunkBase>()

        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.forEach { chunk ->
                chunk.next?.let { next -> linkedChunks += next }
                chunk.instructions.forEach {
                    if(it.branchTarget==null) {
                        it.labelSymbol?.let { label -> allLabeledChunks[label]?.let { cc -> linkedChunks += cc } }
                    } else {
                        linkedChunks += it.branchTarget!!
                    }
                }
                if (chunk.label == "main.start")
                    linkedChunks += chunk
            }
        }

        return removeUnlinkedChunks(linkedChunks)
    }

    private fun removeUnlinkedChunks(
        linkedChunks: MutableSet<IRCodeChunkBase>
    ): Int {
        var numRemoved = 0
        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.withIndex().reversed().forEach { (index, chunk) ->
                if (chunk !in linkedChunks) {
                    if (chunk === sub.chunks[0]) {
                        when(chunk) {
                            is IRCodeChunk -> {
                                if (chunk.isNotEmpty()) {
                                    // don't remove the first chunk of the sub itself because it has to have the name of the sub as label
                                    chunk.instructions.clear()
                                    numRemoved++
                                }
                            }
                            is IRInlineAsmChunk, is IRInlineBinaryChunk -> {
                                sub.chunks[index] = IRCodeChunk(chunk.label, chunk.next)
                                numRemoved++
                            }
                        }
                    } else {
                        sub.chunks.removeAt(index)
                        numRemoved++
                    }
                }
            }
        }
        return numRemoved
    }
}

==> v8.13.zip: prog8/codeGenIntermediate/src/prog8/codegen/intermediate/IRUnusedCodeRemover.kt <==
package prog8.codegen.intermediate

import prog8.code.core.IErrorReporter
import prog8.code.core.SourceCode.Companion.libraryFilePrefix
import prog8.intermediate.*


class IRUnusedCodeRemover(
    private val irprog: IRProgram,
    private val errors: IErrorReporter
) {
    fun optimize(): Int {
        var numRemoved = removeUnusedSubroutines() + removeUnusedAsmSubroutines()

        // remove empty blocks
        irprog.blocks.reversed().forEach { block ->
            if(block.isEmpty()) {
                irprog.blocks.remove(block)
                irprog.st.removeTree(block.label)
                numRemoved++
            }
        }

        return numRemoved
    }

    private fun removeUnusedSubroutines(): Int {
        val allLabeledChunks = mutableMapOf<String, IRCodeChunkBase>()
        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.forEach { chunk ->
                chunk.label?.let { allLabeledChunks[it] = chunk }
            }
        }

        var numRemoved = removeSimpleUnlinked(allLabeledChunks) + removeUnreachable(allLabeledChunks)
        irprog.blocks.forEach { block ->
            block.children.filterIsInstance<IRSubroutine>().reversed().forEach { sub ->
                if(sub.isEmpty()) {
                    if(!sub.position.file.startsWith(libraryFilePrefix)) {
                        errors.warn("unused subroutine ${sub.label}", sub.position)
                    }
                    block.children.remove(sub)
                    irprog.st.removeTree(sub.label)
                    numRemoved++
                }
            }
        }

        return numRemoved
    }

    private fun removeUnusedAsmSubroutines(): Int {
        val allLabeledAsmsubs = irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRAsmSubroutine>() }
            .associateBy { it.label }

        var numRemoved = removeSimpleUnlinkedAsmsubs(allLabeledAsmsubs)
        irprog.blocks.forEach { block ->
            block.children.filterIsInstance<IRAsmSubroutine>().reversed().forEach { sub ->
                if(sub.isEmpty()) {
                    if(!sub.position.file.startsWith(libraryFilePrefix)) {
                        errors.warn("unused asmsubroutine ${sub.label}", sub.position)
                    }
                    block.children.remove(sub)
                    irprog.st.removeTree(sub.label)
                    numRemoved++
                }
            }
        }

        return numRemoved
    }

    private fun removeSimpleUnlinkedAsmsubs(allSubs: Map<String, IRAsmSubroutine>): Int {
        val linkedAsmSubs = mutableSetOf<IRAsmSubroutine>()

        // TODO: asmsubs in library modules are never removed, we can't really tell here if they're actually being called or not...

        // check if asmsub is called from another asmsub
        irprog.blocks.asSequence().forEach { block ->
            block.children.filterIsInstance<IRAsmSubroutine>().forEach { sub ->
                if (block.forceOutput || block.library)
                    linkedAsmSubs += sub
                if (sub.asmChunk.isNotEmpty()) {
                    allSubs.forEach { (label, asmsub) ->
                        if (sub.asmChunk.assembly.contains(label))
                            linkedAsmSubs += asmsub
                    }
                }
                val inlineAsm = sub.asmChunk.next as? IRInlineAsmChunk
                if(inlineAsm!=null) {
                    allSubs.forEach { (label, asmsub) ->
                        if (inlineAsm.assembly.contains(label))
                            linkedAsmSubs += asmsub
                    }
                }
            }
        }

        // check if asmsub is linked or called from another regular subroutine
        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.forEach { chunk ->
                chunk.instructions.forEach {
                    it.labelSymbol?.let { label -> allSubs[label]?.let { cc -> linkedAsmSubs += cc } }
                    // note: branchTarget can't yet point to another IRAsmSubroutine, so do nothing when it's set
                }
            }
        }

        return removeUnlinkedAsmsubs(linkedAsmSubs)
    }

    private fun removeUnlinkedAsmsubs(linkedAsmSubs: Set<IRAsmSubroutine>): Int {
        var numRemoved = 0
        irprog.blocks.asSequence().forEach { block ->
            block.children.withIndex().reversed().forEach { (index, child) ->
                if(child is IRAsmSubroutine && child !in linkedAsmSubs) {
                    block.children.removeAt(index)
                    numRemoved++
                }
            }
        }
        return numRemoved
    }

    private fun removeUnreachable(allLabeledChunks: MutableMap<String, IRCodeChunkBase>): Int {
        val entrypointSub = irprog.blocks.single { it.label=="main" }.children.single { it is IRSubroutine && it.label=="main.start" }
        val reachable = mutableSetOf((entrypointSub as IRSubroutine).chunks.first())

        fun grow() {
            val new = mutableSetOf<IRCodeChunkBase>()
            reachable.forEach {
                it.next?.let { next -> new += next }
                it.instructions.forEach { instr ->
                    if (instr.branchTarget == null)
                        instr.labelSymbol?.let { label -> allLabeledChunks[label]?.let { chunk -> new += chunk } }
                    else
                        new += instr.branchTarget!!
                }
            }
            reachable += new
        }

        var previousCount = reachable.size
        while(true) {
            grow()
            if(reachable.size<=previousCount)
                break
            previousCount = reachable.size
        }

        return removeUnlinkedChunks(reachable)
    }

    private fun removeSimpleUnlinked(allLabeledChunks: Map<String, IRCodeChunkBase>): Int {
        val linkedChunks = mutableSetOf<IRCodeChunkBase>()

        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.forEach { chunk ->
                chunk.next?.let { next -> linkedChunks += next }
                chunk.instructions.forEach {
                    if(it.branchTarget==null) {
                        it.labelSymbol?.let { label -> allLabeledChunks[label]?.let { cc -> linkedChunks += cc } }
                    } else {
                        linkedChunks += it.branchTarget!!
                    }
                }
                if (chunk.label == "main.start")
                    linkedChunks += chunk
            }
        }

        return removeUnlinkedChunks(linkedChunks)
    }

    private fun removeUnlinkedChunks(
        linkedChunks: Set<IRCodeChunkBase>
    ): Int {
        var numRemoved = 0
        irprog.blocks.asSequence().flatMap { it.children.filterIsInstance<IRSubroutine>() }.forEach { sub ->
            sub.chunks.withIndex().reversed().forEach { (index, chunk) ->
                if (chunk !in linkedChunks) {
                    if (chunk === sub.chunks[0]) {
                        when(chunk) {
                            is IRCodeChunk -> {
                                if (chunk.isNotEmpty()) {
                                    // don't remove the first chunk of the sub itself because it has to have the name of the sub as label
                                    chunk.instructions.clear()
                                    numRemoved++
                                }
                            }
                            is IRInlineAsmChunk, is IRInlineBinaryChunk -> {
                                sub.chunks[index] = IRCodeChunk(chunk.label, chunk.next)
                                numRemoved++
                            }
                        }
                    } else {
                        sub.chunks.removeAt(index)
                        numRemoved++
                    }
                }
            }
        }
        return numRemoved
    }
}
